import gcsfs


class GCSFileSystem(gcsfs.GCSFileSystem):
    def __init__(self, token=None, **kwargs):
        super().__init__(token=token, **kwargs)
        # Temporary bug fix for https://issues.apache.org/jira/browse/ARROW-7867
        # Spark writes an empty file to GCS (to mimic a folder structure) before writing partitioned data
        # Resolve this by ignoring the "empty" file when reading partitioned parquet files
        try:
            # Constant is moved to core module in Pyarrow 10.0.0
            from pyarrow.parquet.core import EXCLUDED_PARQUET_PATHS
        except ImportError:
            # Fallback for Pyarrow versions <10.0.0
            from pyarrow.parquet import EXCLUDED_PARQUET_PATHS
        from pyarrow.parquet import ParquetManifest
        EXCLUDED_PARQUET_PATHS.add("")
        ParquetManifest._should_silently_exclude = (
            GCSFileSystem._should_silently_exclude
        )

    def isdir(self, path):
        info = super(gcsfs.GCSFileSystem, self).info(path)
        return info["type"] == "directory"

    @staticmethod
    # This code is from from pyarrow.parquet.core
    def _should_silently_exclude(self, file_name):
        try:
            # Constant is moved to core module in Pyarrow 10.0.0
            from pyarrow.parquet.core import EXCLUDED_PARQUET_PATHS
        except ImportError:
            # Fallback for Pyarrow versions <10.0.0
            from pyarrow.parquet import EXCLUDED_PARQUET_PATHS

        return (
            file_name.endswith(".crc")
            or file_name.endswith("_$folder$")  # Checksums
            or file_name.startswith(".")  # HDFS directories in S3
            or file_name.startswith("_")  # Hidden files starting with .
            or ".tmp" in file_name  # Hidden files starting with _
            or file_name in EXCLUDED_PARQUET_PATHS  # Temp files
        )



from google.cloud import storage
from gcsfs.retry import HttpError
from dapla.auth import AuthClient
from dapla.gcs import GCSFileSystem
from google.oauth2.credentials import Credentials
import logging
import os
import jwt
import requests
import datetime


logger = logging.getLogger(__name__)


class Doctor:
    """Class of functions that perform checks on Dapla. Checks whether user is authenticated,
    if the keycloak-token is valid and if user has access to GCS. Each method can be run 
    individually or collectively with the 'health' method.
    """
    @staticmethod
    def jupyterhub_auth_valid():
        """Checks wheter user is logged in and authenticated to Jupyterhub.
        """
        print("Checking authentication to JupyterHub...")
        try:
            # Attempt fetching the Jupyterhub user
            AuthClient.fetch_local_user()
        except: 
            return False
        return True


    @staticmethod
    def keycloak_token_valid():
        """Checks whether the keycloak token is valid by attempting to access a keycloak-token protected service
        """
        print("Checking that your keycloak token is valid...")
        algorithms = ["RS256"]

        keycloak_token = AuthClient.fetch_personal_token()

        claims = jwt.decode(keycloak_token, algorithms=algorithms, options={"verify_signature": False})

        if not Doctor._is_token_expired(claims):
            return True
        else:
            return False
  

    @staticmethod
    def _is_token_expired(token):
        try:
            # get the expiry time from the payload
            expiry_time = token['exp']

            # convert the expiry time to a datetime object
            expiry_time = datetime.datetime.fromtimestamp(expiry_time)

            # if the current time is greater than the expiry time, the token is expired
            return datetime.datetime.now() > expiry_time
        except KeyError:
            # if the token does not have an expiry time (exp claim), consider it not expired
            return False
        

    @staticmethod
    def gcs_credentials_valid():
        """Checks whether the users google cloud storage token is valid by accessing a GCS service.
        """
        print("Checking your Google Cloud Storage credentials...")

        # Fetch the google token
        google_token = AuthClient.fetch_google_token()

        try:
            response = requests.get("https://oauth2.googleapis.com/tokeninfo?access_token=%s" % google_token)
        except HttpError as ex:
           
            if str(ex) == "Invalid Credentials, 401":
                return False
            else:
                logger.exception(ex)

        return True


    
    @staticmethod
    def bucket_access():
        """Checks whether user has access to a common google bucket.
        """
        print("Checking that you have access to a common Google Cloud Storage bucket...")

        # Fetch google credentials and create client object
        client = storage.Client(credentials=AuthClient.fetch_google_credentials())

        # Set the bucket that is to be accessed
        if os.environ['CLUSTER_ID'] == "staging-bip-app":
            bucket = 'ssb-staging-dapla-felles-data-delt'
        else:
            bucket = 'ssb-prod-dapla-felles-data-delt'

        try:
            # Attempt to access the bucket
            client.get_bucket(bucket)
        except:
            return False
        else:
            return True


    @classmethod
    def health(cls):
        print("Performing checks...")
        if not cls.jupyterhub_auth_valid():
            print("You are either not logged in or not authenticated to JupyterHub.")
            exit(1)
        if not cls.keycloak_token_valid():
            print("Your keycloak token seems to be expired.")
            exit(1)
        if not cls.gcs_credentials_valid():
            print("Your GCS credentials seem to be invalid. Dapla Doctor could not fetch GCS bucket.")
            exit(1)
        if not cls.bucket_access():
            print("You do not seem to have access to a common GCS bucket.")
            exit(1)
        print("Everything seems to be in order.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dapla', 'dapla.spark']

package_data = \
{'': ['*']}

install_requires = \
['PyJWT>=2.6.0,<3.0.0',
 'gcsfs>=2022.7.1',
 'google-cloud-storage>=2.7.0,<3.0.0',
 'ipython>=8.6.0,<9.0.0',
 'jupyterhub>=3.0.0,<4.0.0',
 'lxml>=4.9.1,<5.0.0',
 'pandas>=1.4.2',
 'pyarrow>=8.0.0',
 'pydantic>=1.9.1',
 'requests>=2.27.1']

setup_kwargs = {
    'name': 'dapla-toolbelt',
    'version': '1.3.9',
    'description': "Python module for use within Jupyterlab notebooks, specifically aimed for Statistics Norway's data platform called Dapla",
    'long_description': '# dapla-toolbelt\n\nPython module for use within Jupyterlab notebooks, specifically aimed for Statistics Norway\'s data platform called \n`Dapla`. It contains support for authenticated access to Google Services such as Google Cloud Storage (GCS) and custom\nDapla services such as [Maskinporten Guardian](https://github.com/statisticsnorway/maskinporten-guardian). The \nauthentication process is based on the [TokenExchangeAuthenticator](https://github.com/statisticsnorway/jupyterhub-extensions/tree/main/TokenExchangeAuthenticator)\nfor Jupyterhub.\n\n[![PyPI version](https://img.shields.io/pypi/v/dapla-toolbelt.svg)](https://pypi.python.org/pypi/dapla-toolbelt/)\n![Unit tests](https://github.com/statisticsnorway/dapla-toolbelt/actions/workflows/unit-tests.yml/badge.svg) \n![Code coverage](https://img.shields.io/endpoint?url=https://gist.githubusercontent.com/bjornandre/73205f2f30335801fa2819c31b3ecf79/raw/pytest-coverage-badge-dapla-toolbelt.json) \n![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)\n[![License](https://img.shields.io/pypi/l/dapla-toolbelt.svg)](https://pypi.python.org/pypi/dapla-toolbelt/)\n\nThese operations are supported:\n* List contents of a bucket\n* Open a file in GCS\n* Copy a file from GCS into local\n* Load a file (CSV, JSON or XML) from GCS into a pandas dataframe\n* Save contents of a data frame into a file (CSV, JSON, XML) in GCS\n\nWhen the user gives the path to a resource, they do not need to give the GCS uri, only the path. \nThis just means users don\'t have to prefix a path with "gs://". \nIt is implicitly understood that all resources accessed with this tool are located in GCS, \nwith the first level of the path being a GCS bucket name.\n\n## Installation\n\n`pip install dapla-toolbelt`\n\n## Usage Examples\n\n``` python\nfrom dapla import FileClient\nfrom dapla import GuardianClient\nimport pandas as pd\n\n# Load data using the Maskinporten Guardian client\nresponse = GuardianClient.call_api("https://data.udir.no/api/kag", "88ace991-7871-4ccc-aaec-8fb6d78ed04e", "udir:datatilssb")\ndata_json = response.json()\n\nraw_data_df = pd.DataFrame(data_json)  # create pandas data frame from json\nraw_data_df.head()  # show first rows of data frame\n\nFileClient.ls("bucket-name/folder")  # list contents of given folder\n\n# Save data into different formats\npath_base = "bucket-name/folder/raw_data"\nFileClient.save_pandas_to_json(raw_data_df, f"{path_base}.json")  # generate json from data frame, and save to given path\nFileClient.save_pandas_to_csv(raw_data_df, f"{path_base}.csv")  # generate csv from data frame, and save to given path\nFileClient.save_pandas_to_xml(raw_data_df, f"{path_base}.xml")  # generate xml from data frame, and save to given path\n\nFileClient.cat(f"{path_base}.json")  # print contents of file\n\n# Load data from different formats\n# All these data frames should contain the same data:\ndf = FileClient.load_json_to_pandas(f"{path_base}.json")  # read json from path and load into pandas data frame\ndf.head()  # show first rows of data frame\ndf = FileClient.load_csv_to_pandas(f"{path_base}.csv")  # read csv from path and load into pandas data frame\ndf.head()  # show first rows of data frame\ndf = FileClient.load_xml_to_pandas(f"{path_base}.xml")  # read xml from path and load into pandas data frame\ndf.head()  # show first rows of data frame\n\n```\n\n## Development\n\n### Prerequisites\n\n- Python >3.8 (3.10 is preferred)\n- Poetry, install via `curl -sSL https://install.python-poetry.org | python3 -`\n\nYou can also execute `make` in the project folder to see available `make` commands.\n\n### Dependency Management\n\nPoetry is used for dependency management. [Poe the Poet](https://github.com/nat-n/poethepoet) is used for running poe tasks within poetry\'s virtualenv. \nUpon cloning this project first install necessary dependencies, then run the tests to verify everything is working.\n\n#### Install all dependencies\n\n```shell\npoetry install\n```\n\n### Add dependencies\n\n#### Main\n\n```shell\npoetry add <python package name>\n```\n\n#### Dev\n\n```shell\npoetry add --group dev <python package name>\n```\n\n### Run tests\n\n```shell\npoetry run poe test\n```\n\n### Run project locally in Jupyter\n\nTo run the project locally in Jupyter run:\n\n```shell\npoetry run poe jupyter\n```\n\nA Jupyter instance should open in your browser. Open and run the cells in the `demo.ipynb` file.\n\n### Bumping version\n\nUse `make` to bump the *patch*, *minor* version or *major* version before creating a pull request to the `main` GIT branch.\nOr run a poe task like this:\n\n```shell\npoetry run poe bump-patch-version\n```\n\nYou can use either `bump-version-patch`, `bump-version-minor`, or `bump-version-major`.\nBumping must be done with a clean git working space, and automatically commits with the new version number.\n\nThen just run `git push origin --tags` to push the changes and trigger the release process.\n\n### Building and releasing\n\nBefore merging your changes into the `main` branch, make sure you have bumped the version like outlined above.\n\nAn automatic release process will build *dapla-toolbelt* upon pull request-creation, merges, and direct commits to the\n`main` GIT branch. It will also release a new version of the package to **pypi.org** automatically when a commit is \ntagged, for example by a GitHub release.\n\n### Building and releasing manually\n\nRun `make build` to build a wheel and a source distribution.\n\nRun `make release-validate` to do all that AND validate it for release.\n\nRun this (replacing <SEMVER> with your current version number) to check the contents of your wheel:\n`tar tzf dist/dapla-toolbelt-<SEMVER>.tar.gz`\n\n#### Test release\n\nYou have to bump the version of the package (see documentation on "Bumping version" above) before releasing, \nbecause even test.pypi.org does not allow re-releases of a previously released version.\n\nRun the following command in order to build, validate, and test package publication by uploading to TestPyPI:\n`make release-test`\n\nYou will have to manually enter a username and password for a user registered to [test.pypi.org](https://test.pypi.org) \nin order for this to work.\n\n#### Production release\n\n**NB: A manual production release should only be done as a last resort**, if the regular CI/CD pipeline \ndoes not work, and it\'s necessary to release regardless.\n\nYou have to bump the version of the package (see documentation on "Bumping version" above) to something \ndifferent from the last release before releasing.\n\nIn order to publish a new version of the package to PyPI for real, run `make release`. \nAuthenticate by manually entering your [pypi.org](https://pypi.org) username and password. \n',
    'author': 'Statistics Norway',
    'author_email': 'stat-dev@ssb.no',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/statisticsnorway/dapla-toolbelt',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
